#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars import (
    _CustomBuddhistCalendar,
    _CustomChineseCalendar,
    _CustomHinduCalendar,
    _CustomIslamicCalendar,
)
from holidays.calendars.gregorian import (
    JAN,
    FEB,
    MAR,
    APR,
    MAY,
    JUN,
    JUL,
    AUG,
    SEP,
    OCT,
    NOV,
    DEC,
    MON,
    FRI,
    SAT,
    SUN,
)
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import (
    BuddhistCalendarHolidays,
    ChineseCalendarHolidays,
    ChristianHolidays,
    HinduCalendarHolidays,
    InternationalHolidays,
    IslamicHolidays,
)


class Malaysia(
    HolidayBase,
    BuddhistCalendarHolidays,
    ChineseCalendarHolidays,
    ChristianHolidays,
    HinduCalendarHolidays,
    InternationalHolidays,
    IslamicHolidays,
):
    country = "MY"
    special_holidays = {
        # The years 1955 1959 1995 seems to have the elections
        # one weekday but I am not sure if they were marked as
        # holidays.
        1999: (NOV, 29, "Malaysia General Election Holiday"),
        2018: (MAY, 9, "Malaysia General Election Holiday"),
        2019: (JUL, 30, "Installation of New King"),
    }
    subdivisions = (
        "JHR",
        "KDH",
        "KTN",
        "KUL",
        "LBN",
        "MLK",
        "NSN",
        "PHG",
        "PJY",
        "PLS",
        "PNG",
        "PRK",
        "SBH",
        "SGR",
        "SWK",
        "TRG",
    )

    def __init__(self, *args, **kwargs) -> None:
        """
        An subclass of :py:class:`HolidayBase` representing public holidays in
        Malaysia.

        If ``subdiv`` for a state is not supplied, only nationwide holidays are
        returned. The following ``subdiv`` state codes are used (ISO 3166-2
        subdivision codes are not yet supported):

        - JHR: Johor
        - KDH: Kedah
        - KTN: Kelantan
        - MLK: Melaka
        - NSN: Negeri Sembilan
        - PHG: Pahang
        - PRK: Perak
        - PLS: Perlis
        - PNG: Pulau Pinang
        - SBH: Sabah
        - SWK: Sarawak
        - SGR: Selangor
        - TRG: Terengganu
        - KUL: FT Kuala Lumpur
        - LBN: FT Labuan
        - PJY: FT Putrajaya

        Limitations:

        - Prior to 2021: holidays are not accurate.
        - 2027 and later: Thaipusam dates are are estimated, and so denoted.

        Reference: `Wikipedia
        <https://en.wikipedia.org/wiki/Public_holidays_in_Malaysia>`__

        Country created by: `Eden <https://github.com/jusce17>`__

        Country maintained by: `Mike Borsetti <https://github.com/mborsetti>`__

        See parameters and usage in :py:class:`HolidayBase`.
        """
        BuddhistCalendarHolidays.__init__(
            self, calendar=MalaysiaBuddhistCalendar(), show_estimated=True
        )
        ChineseCalendarHolidays.__init__(
            self, calendar=MalaysiaChineseCalendar(), show_estimated=True
        )
        ChristianHolidays.__init__(self)
        HinduCalendarHolidays.__init__(self, calendar=MalaysiaHinduCalendar())
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self, calendar=MalaysiaIslamicCalendar())
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)
        observed_dates = set()

        # Chinese New Year (one day in the States of Kelantan and Terengganu,
        # two days in the other States).
        observed_dates.add(self._add_chinese_new_years_day("Chinese New Year"))
        # The second day of Chinese New Year is not a federal holiday in
        # Kelantan and Terengganu. However, it is gazetted as a state holiday
        # in both states, effectively making it a nationwide holiday.
        observed_dates.add(self._add_chinese_new_years_day_two("Chinese New Year Holiday"))

        # Vesak Day.
        observed_dates.add(self._add_vesak_may("Vesak Day"))

        # Labour Day.
        observed_dates.add(self._add_labor_day("Labour Day"))

        # Birthday of [His Majesty] the Yang di-Pertuan Agong.
        if year <= 2017:
            hol_date = self._get_nth_weekday_of_month(1, SAT, JUN)
        elif year == 2018:
            hol_date = date(2018, SEP, 9)
        elif year == 2020:
            # https://www.nst.com.my/news/nation/2020/03/571660/agongs-birthday-moved-june-6-june-8
            hol_date = date(2020, JUN, 8)
        else:
            hol_date = self._get_nth_weekday_of_month(1, MON, JUN)
        observed_dates.add(self._add_holiday("Birthday of SPB Yang di-Pertuan Agong", hol_date))

        # Hari Kebangsaan or National Day.
        observed_dates.add(self._add_holiday("National Day", AUG, 31))

        # Malaysia Day.
        if year >= 2010:
            observed_dates.add(self._add_holiday("Malaysia Day", SEP, 16))

        # Deepavali (Diwali).
        if self.subdiv != "SWK":
            observed_dates.add(self._add_diwali("Deepavali"))

        # Christmas day.
        observed_dates.add(self._add_christmas_day("Christmas Day"))

        # Birthday of the Prophet Muhammad (s.a.w.).
        # a.k.a. Hari Keputeraan Nabi Muhammad (Sabah Act)
        observed_dates.update(
            self._add_mawlid_day("Maulidur Rasul (Birthday of the Prophet Muhammad)")
        )

        # Hari Raya Puasa (2 days).
        # aka Eid al-Fitr;
        # exact date of observance is announced yearly
        name = "Hari Raya Puasa"
        observed_dates.update(self._add_eid_al_fitr_day(name))
        observed_dates.update(self._add_eid_al_fitr_day_two(f"Second day of {name}"))

        # Arafat Day.
        if self.subdiv == "TRG":
            observed_dates.update(self._add_arafah_day("Arafat Day"))

        # Hari Raya Haji.
        name = "Hari Raya Haji"
        observed_dates.update(self._add_eid_al_adha_day(name))
        if self.subdiv in {"KDH", "KTN", "PLS", "TRG"}:
            observed_dates.update(self._add_eid_al_adha_day_two(name))

        # New Year's Day
        if self.subdiv in {
            "KUL",
            "LBN",
            "MLK",
            "NSN",
            "PHG",
            "PJY",
            "PNG",
            "PRK",
            "SBH",
            "SGR",
            "SWK",
        }:
            observed_dates.add(self._add_new_years_day("New Year's Day"))

        # Isra and Mi'raj.
        if self.subdiv in {"KDH", "NSN", "PLS"} or (self.subdiv == "TRG" and year >= 2020):
            observed_dates.update(self._add_isra_and_miraj_day("Isra and Mi'raj"))

        # Beginning of Ramadan.
        if self.subdiv in {"JHR", "KDH", "MLK"}:
            observed_dates.update(self._add_ramadan_beginning_day("Beginning of Ramadan"))

        # Nuzul Al-Quran Day.
        if self.subdiv in {
            "KTN",
            "KUL",
            "LBN",
            "PHG",
            "PJY",
            "PLS",
            "PNG",
            "PRK",
            "SGR",
            "TRG",
        }:
            observed_dates.update(self._add_nuzul_al_quran_day("Nuzul Al-Quran Day"))

        # Thaipusam.
        if self.subdiv in {"JHR", "KUL", "NSN", "PJY", "PNG", "PRK", "SGR"}:
            observed_dates.add(self._add_thaipusam("Thaipusam"))

        # Federal Territory Day.
        if self.subdiv in {"KUL", "LBN", "PJY"} and year >= 1974:
            observed_dates.add(self._add_holiday("Federal Territory Day", FEB, 1))

        # State holidays (single state)

        if self.subdiv == "MLK":
            if year >= 1989:
                observed_dates.add(
                    self._add_holiday("Declaration of Malacca as a Historical City", APR, 15)
                )

            dt = (
                date(year, AUG, 24)
                if year >= 2020
                else self._get_nth_weekday_of_month(2, FRI, OCT)
            )
            observed_dates.add(self._add_holiday("Birthday of the Governor of Malacca", dt))

        elif self.subdiv == "NSN" and year >= 2009:
            observed_dates.add(
                self._add_holiday("Birthday of the Sultan of Negeri Sembilan", JAN, 14)
            )

        elif self.subdiv == "PHG" and year >= 1975:
            observed_dates.add(
                self._add_holiday("Hari Hol of Pahang", MAY, 22 if year >= 2021 else 7)
            )

            hol_date = (JUL, 30) if year >= 2019 else (OCT, 24)
            observed_dates.add(self._add_holiday("Birthday of the Sultan of Pahang", *hol_date))

        elif self.subdiv == "PNG":
            if year >= 2009:
                observed_dates.add(self._add_holiday("George Town Heritage Day", JUL, 7))

            observed_dates.add(
                self._add_holiday(
                    "Birthday of the Governor of Penang",
                    self._get_nth_weekday_of_month(2, SAT, JUL),
                )
            )

        elif self.subdiv == "PLS" and year >= 2000:
            hol_date = (JUL, 17) if year >= 2018 else (MAY, 17)
            observed_dates.add(self._add_holiday("Birthday of The Raja of Perlis", *hol_date))

        elif self.subdiv == "SGR":
            observed_dates.add(self._add_holiday("Birthday of The Sultan of Selangor", DEC, 11))

        elif self.subdiv == "SWK":
            # Dayak Festival Day (the first day of June) and the following day.
            observed_dates.add(self._add_holiday("Gawai Dayak", JUN, 1))
            observed_dates.add(self._add_holiday("Gawai Dayak (Second day)", JUN, 2))

            # Birthday of Tuan Yang Terutama Yang di-Pertua Negeri Sarawak
            # (the second Saturday of October).
            observed_dates.add(
                self._add_holiday(
                    "Birthday of the Governor of Sarawak",
                    self._get_nth_weekday_of_month(2, SAT, OCT),
                )
            )

            # Sarawak Independence Day
            if year >= 2017:
                observed_dates.add(self._add_holiday("Sarawak Day", JUL, 22))

        elif self.subdiv == "TRG" and year >= 2000:
            observed_dates.add(
                self._add_holiday(
                    "Anniversary of the Installation of the Sultan of Terengganu", MAR, 4
                )
            )

            observed_dates.add(self._add_holiday("Birthday of the Sultan of Terengganu", APR, 26))

        # Check for holidays that fall on a Sunday and
        # implement Section 3 of Malaysian Holidays Act:
        # "if any day specified in the Schedule falls on
        # Sunday then the day following shall be a public
        # holiday and if such day is already a public holiday,
        # then the day following shall be a public holiday"
        # In Johor and Kedah it's Friday -> Sunday,
        # in Kelantan and Terengganu it's Saturday -> Sunday
        if self.observed:
            weekday_observed_days = (
                (FRI, +2)
                if self.subdiv in {"JHR", "KDH"}
                else ((SAT, +1) if self.subdiv in {"KTN", "TRG"} else (SUN, +1))
            )
            observed_dates.difference_update({None})
            for hol_date in sorted(observed_dates):
                if hol_date.weekday() != weekday_observed_days[0]:
                    continue
                in_lieu_date = hol_date + td(days=weekday_observed_days[1])
                while in_lieu_date in observed_dates:
                    in_lieu_date += td(days=+1)
                for hol_name in self.get_list(hol_date):
                    self._add_holiday(f"{hol_name} [In lieu]", in_lieu_date)
                observed_dates.add(in_lieu_date)

            # special cases (observed from previuos year)
            if year == 2007 and self.subdiv not in {
                "JHR",
                "KDH",
                "KTN",
                "TRG",
            }:
                self._add_holiday("Hari Raya Haji [In lieu]", JAN, 2)

            if year == 2007 and self.subdiv == "TRG":
                self._add_holiday("Arafat Day [In lieu]", JAN, 2)

        # The last two days in May (Pesta Kaamatan).
        # (Sarawak Act)
        # Day following a Sunday is not a holiday
        if self.subdiv in {"LBN", "SBH"}:
            self._add_holiday("Pesta Kaamatan", MAY, 30)
            self._add_holiday("Pesta Kaamatan (Second day)", MAY, 31)

        # Other holidays (decrees etc.)

        # Awal Muharram.
        self._add_islamic_new_year_day("Awal Muharram (Hijri New Year)")

        # Special holidays (states)
        if year == 2021 and self.subdiv in {"KUL", "LBN", "PJY"}:
            self._add_holiday("Malaysia Cup Holiday", DEC, 3)

        if year == 2022 and self.subdiv == "KDH":
            self._add_holiday("Thaipusam", JAN, 18)

        if year == 2022 and self.subdiv in {"JHR", "KDH", "KTN", "TRG"}:
            self._add_holiday("Labour Day Holiday", MAY, 4)

        # Multiple state holidays.
        # Good Friday.
        if self.subdiv in {"SBH", "SWK"}:
            self._add_good_friday("Good Friday")

        # Single state holidays.
        if self.subdiv == "JHR":
            if year >= 2015:
                self._add_holiday("Birthday of the Sultan of Johor", MAR, 23)

            if year >= 2011:
                self._add_hari_hol_johor("Hari Hol of Sultan Iskandar of Johor")

        elif self.subdiv == "KDH" and year >= 2020:
            self._add_holiday(
                "Birthday of The Sultan of Kedah", self._get_nth_weekday_of_month(3, SUN, JUN)
            )

        elif self.subdiv == "KTN" and year >= 2010:
            name = "Birthday of the Sultan of Kelantan"
            self._add_holiday(name, NOV, 11)
            self._add_holiday(name, NOV, 12)

        elif self.subdiv == "PRK":
            # This Holiday used to be on 27th until 2017
            # https://www.officeholidays.com/holidays/malaysia/birthday-of-the-sultan-of-perak
            dt = (
                self._get_nth_weekday_of_month(1, FRI, NOV)
                if year >= 2018
                else date(year, NOV, 27)
            )
            self._add_holiday("Birthday of the Sultan of Perak", dt)

        elif self.subdiv == "SBH":
            self._add_holiday(
                "Birthday of the Governor of Sabah", self._get_nth_weekday_of_month(1, SAT, OCT)
            )

            if year >= 2019:
                self._add_christmas_eve("Christmas Eve")


class MY(Malaysia):
    pass


class MYS(Malaysia):
    pass


class MalaysiaBuddhistCalendar(_CustomBuddhistCalendar):
    VESAK_MAY_DATES = {
        2001: (MAY, 7),
        2002: (MAY, 27),
        2003: (MAY, 15),
        2004: (MAY, 3),
        2005: (MAY, 22),
        2006: (MAY, 12),
        2007: (MAY, 1),
        2008: (MAY, 19),
        2009: (MAY, 9),
        2010: (MAY, 28),
        2011: (MAY, 17),
        2012: (MAY, 5),
        2013: (MAY, 24),
        2014: (MAY, 13),
        2015: (MAY, 3),
        2016: (MAY, 21),
        2017: (MAY, 10),
        2018: (MAY, 29),
        2019: (MAY, 19),
        2020: (MAY, 7),
        2021: (MAY, 26),
        2022: (MAY, 15),
        2023: (MAY, 4),
    }


class MalaysiaChineseCalendar(_CustomChineseCalendar):
    LUNAR_NEW_YEAR_DATES = {
        2001: (JAN, 24),
        2002: (FEB, 12),
        2003: (FEB, 1),
        2004: (JAN, 22),
        2005: (FEB, 9),
        2006: (JAN, 29),
        2007: (FEB, 18),
        2008: (FEB, 7),
        2009: (JAN, 26),
        2010: (FEB, 14),
        2011: (FEB, 3),
        2012: (JAN, 23),
        2013: (FEB, 10),
        2014: (JAN, 31),
        2015: (FEB, 19),
        2016: (FEB, 8),
        2017: (JAN, 28),
        2018: (FEB, 16),
        2019: (FEB, 5),
        2020: (JAN, 25),
        2021: (FEB, 12),
        2022: (FEB, 1),
        2023: (JAN, 22),
    }


class MalaysiaHinduCalendar(_CustomHinduCalendar):
    DIWALI_DATES = {
        2001: (NOV, 14),
        2002: (NOV, 3),
        2003: (OCT, 23),
        2004: (NOV, 11),
        2005: (NOV, 1),
        2006: (OCT, 21),
        2007: (NOV, 8),
        2008: (OCT, 27),
        2009: (OCT, 17),
        2010: (NOV, 5),
        2011: (OCT, 26),
        2012: (NOV, 13),
        2013: (NOV, 2),
        2014: (OCT, 22),
        2015: (NOV, 10),
        2016: (OCT, 29),
        2017: (OCT, 18),
        2018: (NOV, 6),
        2019: (OCT, 27),
        2020: (NOV, 14),
        2021: (NOV, 4),
        2022: (OCT, 24),
        2023: (NOV, 12),
    }

    THAIPUSAM_DATES = {
        2018: (JAN, 31),
        2019: (JAN, 21),
        2020: (FEB, 8),
        2021: (JAN, 28),
        2022: (JAN, 18),
        2023: (FEB, 5),
        2024: (JAN, 25),
        2025: (FEB, 11),
        2026: (FEB, 1),
        2027: (JAN, 22),
    }


class MalaysiaIslamicCalendar(_CustomIslamicCalendar):
    EID_AL_ADHA_DATES = {
        2001: (MAR, 6),
        2002: (FEB, 23),
        2003: (FEB, 12),
        2004: (FEB, 2),
        2005: (JAN, 21),
        2006: ((JAN, 10), (DEC, 31)),
        2007: (DEC, 20),
        2008: (DEC, 9),
        2009: (NOV, 28),
        2010: (NOV, 17),
        2011: (NOV, 7),
        2012: (OCT, 26),
        2013: (OCT, 15),
        2014: (OCT, 5),
        2015: (SEP, 24),
        2016: (SEP, 12),
        2017: (SEP, 1),
        2018: (AUG, 22),
        2019: (AUG, 11),
        2020: (JUL, 31),
        2021: (JUL, 20),
        2022: (JUL, 10),
        2023: (JUN, 29),
    }

    EID_AL_FITR_DATES = {
        2001: (DEC, 17),
        2002: (DEC, 6),
        2003: (NOV, 26),
        2004: (NOV, 14),
        2005: (NOV, 3),
        2006: (OCT, 24),
        2007: (OCT, 13),
        2008: (OCT, 1),
        2009: (SEP, 20),
        2010: (SEP, 10),
        2011: (AUG, 31),
        2012: (AUG, 19),
        2013: (AUG, 8),
        2014: (JUL, 28),
        2015: (JUL, 17),
        2016: (JUL, 6),
        2017: (JUN, 25),
        2018: (JUN, 15),
        2019: (JUN, 5),
        2020: (MAY, 24),
        2021: (MAY, 13),
        2022: (MAY, 2),
        2023: (APR, 22),
    }

    HARI_HOL_JOHOR_DATES = {
        2011: (JAN, 12),
        2012: ((JAN, 1), (DEC, 20)),
        2013: (DEC, 10),
        2014: (NOV, 29),
        2015: (NOV, 19),
        2016: (NOV, 7),
        2017: (OCT, 27),
        2018: (OCT, 15),
        2019: (OCT, 5),
        2020: (SEP, 24),
        2021: (SEP, 13),
        2022: (SEP, 3),
    }

    HIJRI_NEW_YEAR_DATES = {
        2001: (MAR, 26),
        2002: (MAR, 15),
        2003: (MAR, 5),
        2004: (FEB, 22),
        2005: (FEB, 10),
        2006: (JAN, 31),
        2007: (JAN, 20),
        2008: ((JAN, 10), (DEC, 29)),
        2009: (DEC, 18),
        2010: (DEC, 8),
        2011: (NOV, 27),
        2012: (NOV, 15),
        2013: (NOV, 5),
        2014: (OCT, 25),
        2015: (OCT, 14),
        2016: (OCT, 2),
        2017: (SEP, 22),
        2018: (SEP, 11),
        2019: (SEP, 1),
        2020: (AUG, 20),
        2021: (AUG, 10),
        2022: (JUL, 30),
    }

    ISRA_AND_MIRAJ_DATES = {
        2001: (OCT, 15),
        2002: (OCT, 4),
        2003: (SEP, 24),
        2004: (SEP, 12),
        2005: (SEP, 1),
        2006: (AUG, 22),
        2007: (AUG, 11),
        2008: (JUL, 31),
        2009: (JUL, 20),
        2010: (JUL, 9),
        2011: (JUN, 29),
        2012: (JUN, 17),
        2013: (JUN, 6),
        2014: (MAY, 27),
        2015: (MAY, 16),
        2016: (MAY, 5),
        2017: (APR, 24),
        2018: (APR, 14),
        2019: (APR, 3),
        2020: (MAR, 22),
        2021: (MAR, 11),
        2022: (MAR, 1),
        2023: (FEB, 18),
    }

    MAWLID_DATES = {
        2001: (JUN, 4),
        2002: (MAY, 24),
        2003: (MAY, 14),
        2004: (MAY, 2),
        2005: (APR, 21),
        2006: (APR, 11),
        2007: (MAR, 31),
        2008: (MAR, 20),
        2009: (MAR, 9),
        2010: (FEB, 26),
        2011: (FEB, 16),
        2012: (FEB, 5),
        2013: (JAN, 24),
        2014: (JAN, 14),
        2015: ((JAN, 3), (DEC, 24)),
        2016: (DEC, 12),
        2017: (DEC, 1),
        2018: (NOV, 20),
        2019: (NOV, 9),
        2020: (OCT, 29),
        2021: (OCT, 19),
        2022: (OCT, 10),
    }
    NUZUL_AL_QURAN_DATES = {
        2001: (DEC, 3),
        2002: (NOV, 22),
        2003: (NOV, 12),
        2004: (NOV, 1),
        2005: (OCT, 21),
        2006: (OCT, 10),
        2007: (SEP, 29),
        2008: (SEP, 18),
        2009: (SEP, 7),
        2010: (AUG, 27),
        2011: (AUG, 17),
        2012: (AUG, 5),
        2013: (JUL, 25),
        2014: (JUL, 15),
        2015: (JUL, 4),
        2016: (JUN, 22),
        2017: (JUN, 12),
        2018: (JUN, 2),
        2019: (MAY, 22),
        2020: (MAY, 10),
        2021: (APR, 29),
        2022: (APR, 19),
        2023: (APR, 8),
    }

    RAMADAN_BEGINNING_DATES = {
        2001: (NOV, 17),
        2002: (NOV, 6),
        2003: (OCT, 27),
        2004: (OCT, 16),
        2005: (OCT, 5),
        2006: (SEP, 24),
        2007: (SEP, 13),
        2008: (SEP, 2),
        2009: (AUG, 22),
        2010: (AUG, 11),
        2011: (AUG, 1),
        2012: (JUL, 20),
        2013: (JUL, 9),
        2014: (JUN, 29),
        2015: (JUN, 18),
        2016: (JUN, 7),
        2017: (MAY, 27),
        2018: (MAY, 17),
        2019: (MAY, 6),
        2020: (APR, 24),
        2021: (APR, 13),
        2022: (APR, 3),
        2023: (MAR, 23),
    }



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import JAN, FEB, MAR, MAY, AUG, SEP, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Namibia(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://www.officeholidays.com/countries/namibia
    https://www.timeanddate.com/holidays/namibia/

    """

    country = "NA"
    special_holidays = {
        # https://gazettes.africa/archive/na/1999/na-government-gazette-dated-1999-11-22-no-2234.pdf
        1999: (DEC, 31, "Y2K changeover"),
        2000: (JAN, 3, "Y2K changeover"),
    }

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date) -> None:
        # https://tinyurl.com/lacorg5835
        # As of 1991/2/1, whenever a public holiday falls on a Sunday,
        # it rolls over to the monday, unless that monday is already
        # a public holiday.
        if self.observed and self._is_sunday(dt) and dt >= date(1991, FEB, 1):
            self._add_holiday("%s (Observed)" % self[dt], dt + td(days=+1))

    def _populate(self, year):
        if year <= 1989:
            return None

        super()._populate(year)

        # New Year's Day.
        self._add_observed(self._add_new_years_day("New Year's Day"))

        # Independence Day.
        self._add_observed(self._add_holiday("Independence Day", MAR, 21))

        # Good Friday.
        self._add_good_friday("Good Friday")

        # Easter Monday.
        self._add_easter_monday("Easter Monday")

        # Workers' Day.
        self._add_observed(self._add_labor_day("Workers' Day"))

        # Cassinga Day.
        self._add_observed(self._add_holiday("Cassinga Day", MAY, 4))

        # Africa Day.
        self._add_observed(self._add_africa_day("Africa Day"))

        # Ascension Day.
        self._add_ascension_thursday("Ascension Day")

        # Heroes' Day.
        self._add_observed(self._add_holiday("Heroes' Day", AUG, 26))

        # http://www.lac.org.na/laws/2004/3348.pdf
        self._add_observed(
            self._add_holiday(
                "Day of the Namibian Women and International Human Rights Day"
                if year >= 2005
                else "International Human Rights Day",
                SEP,
                10,
            )
        )

        # Christmas Day.
        self._add_christmas_day("Christmas Day")

        # Family Day.
        self._add_observed(self._add_christmas_day_two("Family Day"))


class NA(Namibia):
    pass


class NAM(Namibia):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)
#  Copyright: Kateryna Golovanova <kate@kgthreads.com>, 2022

from gettext import gettext as tr

from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Liechtenstein(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Liechtenstein holidays.
    See https://en.wikipedia.org/wiki/Public_holidays_in_Liechtenstein
    for details.
    """

    country = "LI"
    default_language = "de"
    supported_languages = ("de", "en_US", "uk")

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Neujahr"))

        # Saint Berchtold's Day.
        self._add_new_years_day_two(tr("Berchtoldstag"))

        # Epiphany.
        self._add_epiphany_day(tr("Heilige Drei Könige"))

        # Candlemas.
        self._add_candlemas(tr("Mariä Lichtmess"))

        # Shrove Tuesday.
        self._add_carnival_tuesday(tr("Fasnachtsdienstag"))

        # Saint Joseph's Day.
        self._add_saint_josephs_day(tr("Josefstag"))

        # Good Friday.
        self._add_good_friday(tr("Karfreitag"))

        # Easter Sunday.
        self._add_easter_sunday(tr("Ostersonntag"))

        # Easter Monday.
        self._add_easter_monday(tr("Ostermontag"))

        # Labor Day.
        self._add_labor_day(tr("Tag der Arbeit"))

        # Ascension Day.
        self._add_ascension_thursday(tr("Auffahrt"))

        # Whit Sunday.
        self._add_whit_sunday(tr("Pfingstsonntag"))

        # Whit Monday.
        self._add_whit_monday(tr("Pfingstmontag"))

        # Corpus Christi.
        self._add_corpus_christi_day(tr("Fronleichnam"))

        # National Day.
        self._add_assumption_of_mary_day(tr("Staatsfeiertag"))

        # Nativity of Mary.
        self._add_nativity_of_mary_day(tr("Mariä Geburt"))

        # All Saints Day.
        self._add_all_saints_day(tr("Allerheiligen"))

        # Immaculate Conception.
        self._add_immaculate_conception_day(tr("Mariä Empfängnis"))

        # Christmas Eve.
        self._add_christmas_eve(tr("Heiligabend"))

        # Christmas Day.
        self._add_christmas_day(tr("Weihnachten"))

        # St. Stephen's Day.
        self._add_christmas_day_two(tr("Stefanstag"))

        # New Year's Eve.
        self._add_new_years_eve(tr("Silvester"))


class LI(Liechtenstein):
    pass


class LIE(Liechtenstein):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td

from holidays.calendars.gregorian import APR, SEP
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChineseCalendarHolidays, InternationalHolidays


class Vietnam(HolidayBase, ChineseCalendarHolidays, InternationalHolidays):
    """
    https://publicholidays.vn/
    http://vbpl.vn/TW/Pages/vbpqen-toanvan.aspx?ItemID=11013 Article.115
    https://www.timeanddate.com/holidays/vietnam/
    """

    country = "VN"

    def __init__(self, *args, **kwargs):
        ChineseCalendarHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)
        observed_dates = set()

        # New Year's Day
        observed_dates.add(self._add_new_years_day("International New Year's Day"))

        # Lunar New Year
        self._add_chinese_new_years_eve("Vietnamese New Year's Eve")
        self._add_chinese_new_years_day("Vietnamese New Year")
        self._add_chinese_new_years_day_two("The second day of Tet Holiday")
        self._add_chinese_new_years_day_three("The third day of Tet Holiday")
        self._add_chinese_new_years_day_four("The forth day of Tet Holiday")
        self._add_chinese_new_years_day_five("The fifth day of Tet Holiday")

        # Vietnamese Kings' Commemoration Day
        # https://en.wikipedia.org/wiki/H%C3%B9ng_Kings%27_Festival
        if year >= 2007:
            observed_dates.add(self._add_hung_kings_day("Hung Kings Commemoration Day"))

        # Liberation Day/Reunification Day
        observed_dates.add(self._add_holiday("Liberation Day/Reunification Day", APR, 30))

        # International Labor Day
        observed_dates.add(self._add_labor_day("International Labor Day"))

        # Independence Day
        observed_dates.add(self._add_holiday("Independence Day", SEP, 2))

        if self.observed:
            for dt in sorted(observed_dates):
                if not self._is_weekend(dt):
                    continue
                next_workday = dt + td(days=+1)
                while self._is_weekend(next_workday) or next_workday in observed_dates:
                    next_workday += td(days=+1)
                observed_dates.add(self._add_holiday(f"{self[dt]} (Observed)", next_workday))


class VN(Vietnam):
    pass


class VNM(Vietnam):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import MAY, JUL
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Montenegro(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
      - https://en.wikipedia.org/wiki/Public_holidays_in_Montenegro
      - https://me.usembassy.gov/holiday-calendar/
      - https://publicholidays.eu/montenegro/2023-dates/
      - https://www.officeholidays.com/countries/montenegro/2023
    """

    country = "ME"

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, calendar=JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date, days: int = +1) -> None:
        if self.observed and self._is_sunday(dt):
            self._add_holiday("%s (Observed)" % self[dt], dt + td(days))

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        name = "New Year's Day"
        self._add_observed(self._add_new_years_day(name), days=+2)
        self._add_observed(self._add_new_years_day_two(name))

        # Orthodox Christmas Eve.
        self._add_christmas_eve("Orthodox Christmas Eve")

        # Orthodox Christmas.
        self._add_christmas_day("Orthodox Christmas")

        # Labour Day.
        name = "Labour Day"
        self._add_observed(self._add_labor_day(name), days=+2)
        self._add_observed(self._add_labor_day_two(name))

        # Good Friday.
        self._add_good_friday("Orthodox Good Friday")

        # Easter Sunday.
        self._add_easter_sunday("Orthodox Easter Sunday")

        # Easter Monday.
        self._add_easter_monday("Orthodox Easter Monday")

        # Independence Day.
        name = "Independence Day"
        may_21 = self._add_holiday(name, MAY, 21)
        self._add_observed(may_21, days=+2)
        self._add_observed(self._add_holiday(name, may_21 + td(days=+1)))

        # Statehood Day.
        name = "Statehood Day"
        jul_13 = self._add_holiday(name, JUL, 13)
        self._add_observed(jul_13, days=+2)
        self._add_observed(self._add_holiday(name, jul_13 + td(days=+1)))


class ME(Montenegro):
    pass


class MNE(Montenegro):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)


from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, MAR, JUL, AUG, NOV
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import IslamicHolidays, InternationalHolidays


class Morocco(HolidayBase, InternationalHolidays, IslamicHolidays):
    """
    Moroccan holidays
    Note that holidays falling on a sunday is "lost",
    it will not be moved to another day to make up for the collision.

    Holidays after 2020: the following four moving date holidays whose exact
    date is announced yearly are estimated (and so denoted):
    - Eid El Fetr
    - Eid El Adha
    - 1er Moharram
    - Aid al Mawlid Annabawi
    Primary sources:
    https://fr.wikipedia.org/wiki/F%C3%AAtes_et_jours_f%C3%A9ri%C3%A9s_au_Maroc
    https://www.mmsp.gov.ma/fr/pratiques.aspx?id=38
    """

    country = "MA"
    default_language = "ar"
    # Estimated label.
    estimated_label = tr("(تقدير*) *%s")
    supported_languages = ("ar", "en_US")

    def __init__(self, *args, **kwargs):
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        self._add_new_years_day(tr("رأس السنة الميلادية"))

        if year >= 1945:
            # Independence Manifesto Day post 1944
            self._add_holiday(tr("ذكرى تقديم وثيقة الاستقلال"), JAN, 11)

        # In May 2023, Morocco recognised Berber New Year as official holiday
        # http://www.diplomatie.ma/en/statement-royal-office-12
        if year >= 2024:
            # Amazigh New Year
            self._add_holiday(tr("رأس السنة الأمازيغية"), JAN, 13)

        # Labor day
        self._add_labor_day(tr("عيد العمال"))

        if year >= 2001:
            dt = (JUL, 30)
        elif year >= 1963:
            dt = (MAR, 3)
        else:
            dt = (NOV, 18)
        # Throne day
        self._add_holiday(tr("عيد العرش"), *dt)

        # Oued Ed-Dahab Day
        self._add_holiday(tr("ذكرى استرجاع إقليم وادي الذهب"), AUG, 14)

        # Revolution Day
        self._add_holiday(tr("ذكرى ثورة الملك و الشعب"), AUG, 20)

        # Youth day
        self._add_holiday(tr("عيد الشباب"), *((AUG, 21) if year >= 2001 else (JUL, 9)))

        if year >= 1976:
            # Green March
            self._add_holiday(tr("ذكرى المسيرة الخضراء"), NOV, 6)

        if year >= 1957:
            # Independance day
            self._add_holiday(tr("عيد الإستقلال"), NOV, 18)

        # Eid al-Fitr - Feast Festive
        # date of observance is announced yearly, This is an estimate since
        # having the Holiday on Weekend does change the number of days,
        # deceided to leave it since marking a Weekend as a holiday
        # wouldn't do much harm.
        name = tr("عيد الفطر")
        self._add_eid_al_fitr_day(name)
        self._add_eid_al_fitr_day_two(name)

        # Eid al-Adha - Sacrifice Festive
        # date of observance is announced yearly
        name = tr("عيد الأضحى")
        self._add_eid_al_adha_day(name)
        self._add_eid_al_adha_day_two(name)

        # Islamic New Year - (hijari_year, 1, 1)
        self._add_islamic_new_year_day(tr("رأس السنة الهجرية"))

        # Prophet Muhammad's Birthday - (hijari_year, 3, 12)
        name = tr("عيد المولد النبوي")
        self._add_mawlid_day(name)
        self._add_mawlid_day_two(name)


class MA(Morocco):
    pass


class MOR(Morocco):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import JAN, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Panama(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
      - https://en.wikipedia.org/wiki/Public_holidays_in_Panama
      - https://publicholidays.com.pa/
    """

    country = "PA"

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date) -> None:
        if self.observed and self._is_sunday(dt):
            self._add_holiday("%s (Observed)" % self[dt], dt + td(days=+1))

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        self._add_observed(self._add_new_years_day("New Year's Day"))

        # Martyrs' Day
        self._add_observed(self._add_holiday("Martyrs' Day", JAN, 9))

        # Carnival
        self._add_carnival_tuesday("Carnival")

        # Good Friday
        self._add_good_friday("Good Friday")

        # Labour Day
        self._add_observed(self._add_labor_day("Labour Day"))

        # Separation Day
        self._add_holiday("Separation Day", NOV, 3)

        # National Symbols Day
        self._add_holiday("National Symbols Day", NOV, 4)

        # Colon Day
        self._add_holiday("Colon Day", NOV, 5)

        # Los Santos Uprising Day
        self._add_holiday("Los Santos Uprising Day", NOV, 10)

        # Independence Day
        self._add_observed(self._add_holiday("Independence Day", NOV, 28))

        # Mother's Day
        self._add_observed(self._add_holiday("Mother's Day", DEC, 8))

        # National Mourning Day
        if year >= 2022:
            self._add_observed(self._add_holiday("National Mourning Day", DEC, 20))

        # Christmas Day
        self._add_observed(self._add_christmas_day("Christmas Day"))


class PA(Panama):
    pass


class PAN(Panama):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

import warnings
from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import MAR, MAY, JUL, SEP, NOV, DEC, FRI
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class HolidaysMH(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://rmiparliament.org/cms/component/content/article/14-pressrelease/49-important-public-holidays.html?Itemid=101
    https://www.rmiembassyus.org/country-profile#:~:text=national%20holidays
    """

    country = "MH"

    # Special Holidays

    # General Election Day
    election_day = "General Election Day"

    special_holidays = {
        1995: (NOV, 20, election_day),
        1999: (NOV, 22, election_day),
        2003: (NOV, 17, election_day),
        2007: (NOV, 19, election_day),
        2011: (NOV, 21, election_day),
        2015: (NOV, 16, election_day),
        2019: (NOV, 18, election_day),
        2023: (NOV, 20, election_day),
    }

    def _add_observed(self, dt: date) -> None:
        """
        If fall on Sunday, an observed holiday marked with suffix " Holiday" will be added.
        """
        if self.observed and self._is_sunday(dt):
            self._add_holiday("%s Holiday" % self[dt], dt + td(days=+1))

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        if year <= 2019:
            warnings.warn(
                "Years before 2020 are not available for the Marshall Islands (MH).", Warning
            )

        # New Year's Day
        self._add_observed(self._add_new_years_day("New Year's Day"))

        # Nuclear Victims Remembrance Day
        self._add_observed(self._add_holiday("Nuclear Victims Remembrance Day", MAR, 1))

        # Good Friday
        self._add_good_friday("Good Friday")

        # Constitution Day
        self._add_observed(self._add_holiday("Constitution Day", MAY, 1))

        # Fisherman's Day
        self._add_holiday("Fisherman's Day", self._get_nth_weekday_of_month(1, FRI, JUL))

        # Dri-jerbal Day
        self._add_holiday("Dri-jerbal Day", self._get_nth_weekday_of_month(1, FRI, SEP))

        # Manit Day
        self._add_holiday("Manit Day", self._get_nth_weekday_of_month(-1, FRI, SEP))

        # President's Day
        self._add_observed(self._add_holiday("President's Day", NOV, 17))

        # Gospel Day
        self._add_holiday("Gospel Day", self._get_nth_weekday_of_month(1, FRI, DEC))

        # Christmas Day
        name = "Christmas Day"
        if year == 2021:
            # special case
            self._add_holiday(name, DEC, 24)
        else:
            self._add_observed(self._add_christmas_day(name))


class MH(HolidaysMH):
    pass


class MHL(HolidaysMH):
    pass


class MarshallIslands(HolidaysMH):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

import warnings

from holidays.calendars.gregorian import JAN, MAR, APR, MAY, JUN, AUG, OCT, NOV
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays, IslamicHolidays


class India(HolidayBase, ChristianHolidays, InternationalHolidays, IslamicHolidays):
    """
    https://www.india.gov.in/calendar
    https://www.india.gov.in/state-and-ut-holiday-calendar
    https://en.wikipedia.org/wiki/Public_holidays_in_India
    https://www.calendarlabs.com/holidays/india/2021
    https://slusi.dacnet.nic.in/watershedatlas/list_of_state_abbreviation.htm
    https://vahan.parivahan.gov.in/vahan4dashboard/
    """

    country = "IN"
    subdivisions = (
        "AN",  # Andaman and Nicobar Islands.
        "AP",  # Andhra Pradesh.
        "AR",  # Arunachal Pradesh.
        "AS",  # Assam.
        "BR",  # Bihar.
        "CG",  # Chhattisgarh.
        "CH",  # Chandigarh.
        "DD",  # Daman and Diu.
        "DH",  # Dadra and Nagar Haveli.
        "DL",  # Delhi.
        "GA",  # Goa.
        "GJ",  # Gujarat.
        "HP",  # Himachal Pradesh.
        "HR",  # Haryana.
        "JH",  # Jharkhand.
        "JK",  # Jammu and Kashmir.
        "KA",  # Karnataka.
        "KL",  # Kerala.
        "LA",  # Ladakh.
        "LD",  # Lakshadweep.
        "MH",  # Maharashtra.
        "ML",  # Meghalaya.
        "MN",  # Manipur.
        "MP",  # Madhya Pradesh.
        "MZ",  # Mizoram.
        "NL",  # Nagaland.
        "OR",  # Orissa / Odisha (Govt sites (dacnet/vahan) use code "OR").
        "PB",  # Punjab.
        "PY",  # Pondicherry.
        "RJ",  # Rajasthan.
        "SK",  # Sikkim.
        "TN",  # Tamil Nadu.
        "TR",  # Tripura.
        "TS",  # Telangana.
        "UK",  # Uttarakhand.
        "UP",  # Uttar Pradesh.
        "WB",  # West Bengal.
    )

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # Makar Sankranti / Pongal.
        self._add_holiday("Makar Sankranti / Pongal", JAN, 14)

        if year >= 1950:
            # Republic Day.
            self._add_holiday("Republic Day", JAN, 26)

        if year >= 1947:
            # Independence Day.
            self._add_holiday("Independence Day", AUG, 15)

        # Gandhi Jayanti.
        self._add_holiday("Gandhi Jayanti", OCT, 2)

        # Labour Day.
        self._add_labor_day("Labour Day")

        # Directly lifted Diwali and Holi dates from FBProphet from:
        # https://github.com/facebook/prophet/blob/main/python/prophet/hdays.py
        # Warnings kept in place so that users are aware
        if year < 2001 or year > 2030:
            warning_msg = "Diwali and Holi holidays available from 2001 to 2030 only"
            warnings.warn(warning_msg, Warning)

        # https://www.timeanddate.com/holidays/india/diwali
        diwali_dates = {
            2001: (NOV, 14),
            2002: (NOV, 4),
            2003: (OCT, 25),
            2004: (NOV, 12),
            2005: (NOV, 1),
            2006: (OCT, 21),
            2007: (NOV, 9),
            2008: (OCT, 28),
            2009: (OCT, 17),
            2010: (NOV, 5),
            2011: (OCT, 26),
            2012: (NOV, 13),
            2013: (NOV, 3),
            2014: (OCT, 23),
            2015: (NOV, 11),
            2016: (OCT, 30),
            2017: (OCT, 19),
            2018: (NOV, 7),
            2019: (OCT, 27),
            2020: (NOV, 14),
            2021: (NOV, 4),
            2022: (OCT, 24),
            2023: (NOV, 12),
            2024: (NOV, 1),
            2025: (OCT, 20),
            2026: (NOV, 8),
            2027: (OCT, 29),
            2028: (OCT, 17),
            2029: (NOV, 5),
            2030: (OCT, 26),
        }

        # https://www.timeanddate.com/holidays/india/holi
        holi_dates = {
            2001: (MAR, 10),
            2002: (MAR, 29),
            2003: (MAR, 18),
            2004: (MAR, 7),
            2005: (MAR, 26),
            2006: (MAR, 15),
            2007: (MAR, 4),
            2008: (MAR, 22),
            2009: (MAR, 11),
            2010: (MAR, 1),
            2011: (MAR, 20),
            2012: (MAR, 8),
            2013: (MAR, 27),
            2014: (MAR, 17),
            2015: (MAR, 6),
            2016: (MAR, 24),
            2017: (MAR, 13),
            2018: (MAR, 2),
            2019: (MAR, 21),
            2020: (MAR, 10),
            2021: (MAR, 29),
            2022: (MAR, 18),
            2023: (MAR, 8),
            2024: (MAR, 25),
            2025: (MAR, 14),
            2026: (MAR, 4),
            2027: (MAR, 22),
            2028: (MAR, 11),
            2029: (MAR, 1),
            2030: (MAR, 20),
        }

        if year in diwali_dates:
            self._add_holiday("Diwali", *diwali_dates[year])

        if year in holi_dates:
            self._add_holiday("Holi", *holi_dates[year])

        # Islamic holidays.
        # Day of Ashura.
        self._add_ashura_day("Day of Ashura")

        # Birth of the Prophet.
        self._add_mawlid_day("Mawlid")

        # Eid ul-Fitr.
        name = "Eid ul-Fitr"
        self._add_eid_al_fitr_day(name)
        self._add_eid_al_fitr_day_two(name)

        # Eid al-Adha.
        name = "Eid al-Adha"
        self._add_eid_al_adha_day(name)
        self._add_eid_al_adha_day_two(name)

        # Christian holidays.
        self._add_palm_sunday("Palm Sunday")
        self._add_good_friday("Good Friday")
        self._add_easter_sunday("Easter Sunday")
        self._add_whit_sunday("Feast of Pentecost")
        self._add_christmas_day("Christmas Day")

    # Andhra Pradesh.
    def _add_subdiv_ap_holidays(self):
        self._add_holiday("Dr. B. R. Ambedkar's Jayanti", APR, 14)
        self._add_holiday("Andhra Pradesh Foundation Day", NOV, 1)

    # Assam.
    def _add_subdiv_as_holidays(self):
        self._add_holiday("Bihu (Assamese New Year)", APR, 15)

    # Bihar.
    def _add_subdiv_br_holidays(self):
        self._add_holiday("Bihar Day", MAR, 22)
        self._add_holiday("Dr. B. R. Ambedkar's Jayanti", APR, 14)

    # Chhattisgarh.
    def _add_subdiv_cg_holidays(self):
        self._add_holiday("Chhattisgarh Foundation Day", NOV, 1)

    # Gujarat.
    def _add_subdiv_gj_holidays(self):
        self._add_holiday("Uttarayan", JAN, 14)
        self._add_holiday("Gujarat Day", MAY, 1)
        self._add_holiday("Sardar Patel Jayanti", OCT, 31)

    # Haryana.
    def _add_subdiv_hr_holidays(self):
        self._add_holiday("Dr. B. R. Ambedkar's Jayanti", APR, 14)
        self._add_holiday("Haryana Foundation Day", NOV, 1)

    # Karnataka.
    def _add_subdiv_ka_holidays(self):
        self._add_holiday("Karnataka Rajyotsava", NOV, 1)

    # Kerala.
    def _add_subdiv_kl_holidays(self):
        self._add_holiday("Dr. B. R. Ambedkar's Jayanti", APR, 14)
        self._add_holiday("Kerala Foundation Day", NOV, 1)

    # Maharashtra.
    def _add_subdiv_mh_holidays(self):
        self._add_holiday("Dr. B. R. Ambedkar's Jayanti", APR, 14)
        self._add_holiday("Maharashtra Day", MAY, 1)
        self._add_holiday("Dussehra", OCT, 15)

    # Madhya Pradesh.
    def _add_subdiv_mp_holidays(self):
        self._add_holiday("Madhya Pradesh Foundation Day", NOV, 1)

    # Orissa / Odisha.
    def _add_subdiv_or_holidays(self):
        self._add_holiday("Odisha Day (Utkala Dibasa)", APR, 1)
        self._add_holiday("Dr. B. R. Ambedkar's Jayanti", APR, 14)
        self._add_holiday("Maha Vishuva Sankranti / Pana Sankranti", APR, 15)

    # Rajasthan.
    def _add_subdiv_rj_holidays(self):
        self._add_holiday("Rajasthan Day", MAR, 30)
        self._add_holiday("Maharana Pratap Jayanti", JUN, 15)

    # Sikkim.
    def _add_subdiv_sk_holidays(self):
        self._add_holiday("Annexation Day", MAY, 16)

    # Tamil Nadu.
    def _add_subdiv_tn_holidays(self):
        self._add_holiday("Dr. B. R. Ambedkar's Jayanti", APR, 14)
        self._add_holiday("Puthandu (Tamil New Year)", APR, 14)
        self._add_holiday("Puthandu (Tamil New Year)", APR, 15)

    # Telangana.
    def _add_subdiv_ts_holidays(self):
        self._add_holiday("Eid al-Fitr", APR, 6)
        self._add_holiday("Bathukamma Festival", OCT, 6)

    # Uttarakhand.
    def _add_subdiv_uk_holidays(self):
        self._add_holiday("Dr. B. R. Ambedkar's Jayanti", APR, 14)

    # Uttar Pradesh.
    def _add_subdiv_up_holidays(self):
        self._add_holiday("Dr. B. R. Ambedkar's Jayanti", APR, 14)

    # West Bengal.
    def _add_subdiv_wb_holidays(self):
        self._add_holiday("Dr. B. R. Ambedkar's Jayanti", APR, 14)
        self._add_holiday("Pohela Boishakh", APR, 14)
        self._add_holiday("Pohela Boishakh", APR, 15)
        self._add_holiday("Rabindra Jayanti", MAY, 9)


class IN(India):
    pass


class IND(India):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import JUN, JUL, AUG, MON, FRI
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays

from .united_kingdom import UnitedKingdom


class IsleOfMan(UnitedKingdom):
    """Using existing code in UnitedKingdom for now."""

    country = "IM"
    subdivisions = ()  # Override UnitedKingdom subdivisions.

    def __init__(self, *args, **kwargs):  # Override UnitedKingdom __init__().
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        HolidayBase.__init__(self, *args, **kwargs)

    def _populate(self, year: int) -> None:
        super()._populate(year)

        # Easter Monday
        self._add_easter_monday("Easter Monday")

        # Late Summer bank holiday (last Monday in August)
        if year >= 1971:
            self._add_holiday(
                "Late Summer Bank Holiday", self._get_nth_weekday_of_month(-1, MON, AUG)
            )

        # Isle of Man exclusive holidays
        # TT bank holiday (first Friday in June)
        self._add_holiday("TT Bank Holiday", self._get_nth_weekday_of_month(1, FRI, JUN))

        # Tynwald Day
        # Move to the next Monday if falls on a weekend.
        dt = date(year, JUL, 5)
        if self._is_weekend(dt) and year >= 1992:
            dt += td(days=+2 if self._is_saturday(dt) else +1)
        self._add_holiday("Tynwald Day", dt)


class IM(IsleOfMan):
    pass


class IMN(IsleOfMan):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, SUN, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Canada(HolidayBase, ChristianHolidays, InternationalHolidays):
    country = "CA"
    default_language = "en"
    subdivisions = (
        "AB",
        "BC",
        "MB",
        "NB",
        "NL",
        "NS",
        "NT",
        "NU",
        "ON",
        "PE",
        "QC",
        "SK",
        "YT",
    )
    supported_languages = ("ar", "en", "en_US", "fr", "th")

    def __init__(self, *args, **kwargs):
        # Default subdivision to ON; prov for backwards compatibility
        if not kwargs.get("subdiv", kwargs.get("prov")):
            kwargs["subdiv"] = "ON"
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _get_nearest_monday(self, *args) -> date:
        dt = date(self._year, *args)
        return self._get_nth_weekday_from(
            1 if self._is_friday(dt) or self._is_weekend(dt) else -1, MON, dt
        )

    def _add_observed(self, dt: date, include_sat: bool = True, days: int = +1) -> None:
        if not self.observed:
            return None
        if self._is_sunday(dt) or (include_sat and self._is_saturday(dt)):
            self._add_holiday(
                self.tr("%s (Observed)") % self[dt],
                dt + td(days=+2 if self._is_saturday(dt) else days),
            )

    def _populate(self, year):
        if year <= 1866:
            return None

        super()._populate(year)

        # New Year's Day.
        self._add_observed(self._add_new_years_day(tr("New Year's Day")))

        # Good Friday.
        self._add_good_friday(tr("Good Friday"))
        # Easter Monday.
        self._add_easter_monday(tr("Easter Monday"))

        if year <= 1982:
            # Dominion Day.
            self._add_observed(self._add_holiday(tr("Dominion Day"), JUL, 1))

        if self._year >= 1894:
            # Labour Day.
            self._add_holiday(tr("Labour Day"), self._get_nth_weekday_of_month(1, MON, SEP))

        # Christmas Day.
        self._add_observed(self._add_christmas_day(tr("Christmas Day")), days=+2)

        # Boxing Day.
        self._add_observed(self._add_christmas_day_two(tr("Boxing Day")), days=+2)

    def _add_family_day(self):
        # Family Day.
        self._add_holiday(tr("Family Day"), self._get_nth_weekday_of_month(3, MON, FEB))

    def _add_thanksgiving(self):
        if self._year >= 1931:
            dt = (
                # in 1935, Canadian Thanksgiving was moved due to the General
                # Election falling on the second Monday of October
                # http://tiny.cc/can_thkgvg
                date(1935, OCT, 25)
                if self._year == 1935
                else self._get_nth_weekday_of_month(2, MON, OCT)
            )
            # Thanksgiving.
            self._add_holiday(tr("Thanksgiving"), dt)

    def _add_queens_funeral(self):
        if self._year == 2022:
            # Funeral of Queen Elizabeth II.
            self._add_holiday(tr("Funeral of Her Majesty the Queen Elizabeth II"), SEP, 19)

    def _add_subdiv_holidays(self):
        if self._year >= 1983:
            self._add_observed(
                self._add_holiday(
                    (
                        # Memorial Day.
                        tr("Memorial Day")
                        if self.subdiv == "NL"
                        # Canada Day.
                        else tr("Canada Day")
                    ),
                    JUL,
                    1,
                )
            )

        super()._add_subdiv_holidays()

    def _add_subdiv_ab_holidays(self):
        if self._year >= 1990:
            self._add_family_day()

        if self._year >= 1953:
            # Victoria Day.
            self._add_holiday(tr("Victoria Day"), self._get_nth_weekday_from(-1, MON, MAY, 24))

        # https://en.wikipedia.org/wiki/Civic_Holiday#Alberta
        if self._year >= 1974:
            # Heritage Day.
            self._add_holiday(tr("Heritage Day"), self._get_nth_weekday_of_month(1, MON, AUG))

        self._add_thanksgiving()

        if self._year >= 1931:
            # Remembrance Day.
            self._add_remembrance_day(tr("Remembrance Day"))

    def _add_subdiv_bc_holidays(self):
        if self._year >= 2013:
            dt = self._get_nth_weekday_of_month(3 if self._year >= 2019 else 2, MON, FEB)
            self._add_holiday(tr("Family Day"), dt)

        if self._year >= 1953:
            # Victoria Day.
            self._add_holiday(tr("Victoria Day"), self._get_nth_weekday_from(-1, MON, MAY, 24))

        # https://en.wikipedia.org/wiki/Civic_Holiday#British_Columbia
        if self._year >= 1974:
            self._add_holiday(
                # British Columbia Day.
                tr("British Columbia Day"),
                self._get_nth_weekday_of_month(1, MON, AUG),
            )

        self._add_queens_funeral()

        if self._year >= 2023:
            # National Day for Truth and Reconciliation.
            self._add_holiday(tr("National Day for Truth and Reconciliation"), SEP, 30)

        self._add_thanksgiving()

        if self._year >= 1931:
            # Remembrance Day.
            self._add_remembrance_day(tr("Remembrance Day"))

    def _add_subdiv_mb_holidays(self):
        if self._year >= 2008:
            # Louis Riel Day.
            self._add_holiday(tr("Louis Riel Day"), self._get_nth_weekday_of_month(3, MON, FEB))

        if self._year >= 1953:
            # Victoria Day.
            self._add_holiday(tr("Victoria Day"), self._get_nth_weekday_from(-1, MON, MAY, 24))

        if self._year >= 1900:
            name = (
                # Terry Fox Day.
                tr("Terry Fox Day")
                if self._year >= 2015
                # Civic Holiday.
                else tr("Civic Holiday")
            )
            self._add_holiday(name, self._get_nth_weekday_of_month(1, MON, AUG))

        if self._year >= 2021:
            # National Day for Truth and Reconciliation.
            self._add_holiday(tr("National Day for Truth and Reconciliation"), SEP, 30)

        self._add_thanksgiving()

        if self._year >= 1931:
            # Remembrance Day.
            self._add_remembrance_day(tr("Remembrance Day"))

    def _add_subdiv_nb_holidays(self):
        if self._year >= 2018:
            self._add_family_day()

        if self._year >= 1953:
            # Victoria Day.
            self._add_holiday(tr("Victoria Day"), self._get_nth_weekday_from(-1, MON, MAY, 24))

        # https://en.wikipedia.org/wiki/Civic_Holiday#New_Brunswick
        if self._year >= 1900:
            # New Brunswick Day.
            self._add_holiday(tr("New Brunswick Day"), self._get_nth_weekday_of_month(1, MON, AUG))

        self._add_queens_funeral()

        if self._year >= 1931:
            # Remembrance Day.
            self._add_remembrance_day(tr("Remembrance Day"))

    def _add_subdiv_nl_holidays(self):
        if self._year >= 1900:
            # St. Patrick's Day.
            self._add_holiday(tr("St. Patrick's Day"), self._get_nearest_monday(MAR, 17))

        if self._year >= 1990:
            # Nearest Monday to April 23
            # 4/26 is the Monday closer to 4/23 in 2010
            # but the holiday was observed on 4/19? Crazy Newfies!
            dt = date(2010, APR, 19) if self._year == 2010 else self._get_nearest_monday(APR, 23)
            # St. George's Day.
            self._add_holiday(tr("St. George's Day"), dt)

        if self._year >= 1997:
            # Discovery Day.
            self._add_holiday(tr("Discovery Day"), self._get_nearest_monday(JUN, 24))

        self._add_queens_funeral()

        if self._year >= 1931:
            # Remembrance Day.
            self._add_observed(self._add_remembrance_day(tr("Remembrance Day")), include_sat=False)

    def _add_subdiv_ns_holidays(self):
        # http://novascotia.ca/lae/employmentrights/NovaScotiaHeritageDay.asp
        if self._year >= 2015:
            # Heritage Day.
            self._add_holiday(tr("Heritage Day"), self._get_nth_weekday_of_month(3, MON, FEB))

        self._add_queens_funeral()

        if self._year >= 2021:
            # National Day for Truth and Reconciliation.
            self._add_holiday(tr("National Day for Truth and Reconciliation"), SEP, 30)

        if self._year >= 1931:
            # Remembrance Day.
            self._add_observed(self._add_remembrance_day(tr("Remembrance Day")), include_sat=False)

    def _add_subdiv_nt_holidays(self):
        if self._year >= 1953:
            # Victoria Day.
            self._add_holiday(tr("Victoria Day"), self._get_nth_weekday_from(-1, MON, MAY, 24))

        if self._year >= 1996:
            # National Aboriginal Day.
            self._add_holiday(tr("National Aboriginal Day"), JUN, 21)

        if self._year >= 1900:
            # Civic Holiday.
            self._add_holiday(tr("Civic Holiday"), self._get_nth_weekday_of_month(1, MON, AUG))

        self._add_thanksgiving()

        if self._year >= 1931:
            # Remembrance Day.
            self._add_observed(self._add_remembrance_day(tr("Remembrance Day")), include_sat=False)

    def _add_subdiv_nu_holidays(self):
        if self._year >= 1953:
            # Victoria Day.
            self._add_holiday(tr("Victoria Day"), self._get_nth_weekday_from(-1, MON, MAY, 24))

        if self._year >= 2000:
            dt = (APR, 1) if self._year == 2000 else (JUL, 9)
            # Nunavut Day.
            self._add_observed(self._add_holiday(tr("Nunavut Day"), *dt), include_sat=False)

        self._add_thanksgiving()

        if self._year >= 1931:
            # Remembrance Day.
            self._add_remembrance_day(tr("Remembrance Day"))

    def _add_subdiv_on_holidays(self):
        if self._year >= 2008:
            self._add_family_day()

        if self._year >= 1953:
            # Victoria Day.
            self._add_holiday(tr("Victoria Day"), self._get_nth_weekday_from(-1, MON, MAY, 24))

        if self._year >= 1900:
            # Civic Holiday.
            self._add_holiday(tr("Civic Holiday"), self._get_nth_weekday_of_month(1, MON, AUG))

        self._add_thanksgiving()

    def _add_subdiv_pe_holidays(self):
        if self._year >= 2009:
            dt = self._get_nth_weekday_of_month(3 if self._year >= 2010 else 2, MON, FEB)
            # Islander Day.
            self._add_holiday(tr("Islander Day"), dt)

        self._add_queens_funeral()

        if self._year >= 1931:
            # Remembrance Day.
            self._add_observed(self._add_remembrance_day(tr("Remembrance Day")), include_sat=False)

    def _add_subdiv_qc_holidays(self):
        if self._year >= 2003:
            dt = self._get_nth_weekday_from(-1, MON, MAY, 24)
            # National Patriots' Day.
            self._add_holiday(tr("National Patriots' Day"), dt)

        if self._year >= 1925:
            self._add_observed(
                # St. Jean Baptiste Day.
                self._add_saint_johns_day(tr("St. Jean Baptiste Day")),
                include_sat=False,
            )

        self._add_thanksgiving()

    def _add_subdiv_sk_holidays(self):
        if self._year >= 2007:
            self._add_family_day()

        if self._year >= 1953:
            # Victoria Day.
            self._add_holiday(tr("Victoria Day"), self._get_nth_weekday_from(-1, MON, MAY, 24))

        # https://en.wikipedia.org/wiki/Civic_Holiday#Saskatchewan
        if self._year >= 1900:
            # Saskatchewan Day.
            self._add_holiday(tr("Saskatchewan Day"), self._get_nth_weekday_of_month(1, MON, AUG))

        self._add_thanksgiving()

        if self._year >= 1931:
            # Remembrance Day.
            self._add_observed(self._add_remembrance_day(tr("Remembrance Day")), include_sat=False)

    def _add_subdiv_yt_holidays(self):
        # start date?
        # https://www.britannica.com/topic/Heritage-Day-Canadian-holiday
        # Heritage Day was created in 1973
        # by the Heritage Canada Foundation
        # therefore, start date is not earlier than 1974
        # http://heritageyukon.ca/programs/heritage-day
        # https://en.wikipedia.org/wiki/Family_Day_(Canada)#Yukon_Heritage_Day
        # Friday before the last Sunday in February
        if self._year >= 1974:
            self._add_holiday(
                tr("Heritage Day"), self._get_nth_weekday_of_month(-1, SUN, FEB) + td(days=-2)
            )

        if self._year >= 1953:
            # Victoria Day.
            self._add_holiday(tr("Victoria Day"), self._get_nth_weekday_from(-1, MON, MAY, 24))

        if self._year >= 1912:
            # Discovery Day.
            self._add_holiday(tr("Discovery Day"), self._get_nth_weekday_of_month(3, MON, AUG))

        self._add_queens_funeral()

        self._add_thanksgiving()

        if self._year >= 1931:
            # Remembrance Day.
            self._add_remembrance_day(tr("Remembrance Day"))


class CA(Canada):
    pass


class CAN(Canada):
    pass
